@@ -55,6 +55,10 @@ opts = [
                default=3,
                help='Maximum attempts to verify an iSCSI connection is '
                     'active, sleeping 1 second between attempts.'),
+    cfg.IntOpt('partprobe_attempts',
+               default=10,
+               help='Maximum number of attempts to try to read the '
+                    'partition.'),
 ]
 
 CONF = cfg.CONF
@@ -168,12 +172,13 @@ def is_nvme_device(dev):
 def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
                     configdrive_mb, node_uuid, commit=True,
                     boot_option="netboot", boot_mode="bios",
-                    disk_label=None):
+                    disk_label=None, cpu_arch=""):
     """Partition the disk device.
 
     Create partitions for root, swap, ephemeral and configdrive on a
     disk device.
 
+    :param dev: Path for the device to work on.
     :param root_mb: Size of the root partition in mebibytes (MiB).
     :param swap_mb: Size of the swap partition in mebibytes (MiB). If 0,
         no partition will be created.
@@ -189,6 +194,11 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     :param disk_label: The disk label to be used when creating the
         partition table. Valid values are: "msdos", "gpt" or None; If None
         Ironic will figure it out according to the boot_mode parameter.
+    :param cpu_arch: Architecture of the node the disk device belongs to.
+        When using the default value of None, no architecture specific
+        steps will be taken. This default should be used for x86_64. When
+        set to ppc64*, architecture specific steps are taken for booting a
+        partition image locally.
     :returns: A dictionary containing the partition type as Key and partition
         path as Value for the partitions created by this method.
 
@@ -223,11 +233,25 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
                                     boot_flag='boot')
         part_dict['efi system partition'] = part_template % part_num
 
-    if boot_mode == "bios" and boot_option == "local" and disk_label == "gpt":
+    if (boot_mode == "bios" and boot_option == "local" and disk_label == "gpt"
+        and not cpu_arch.startswith('ppc64')):
         part_num = dp.add_partition(CONF.disk_utils.bios_boot_partition_size,
                                     boot_flag='bios_grub')
         part_dict['BIOS Boot partition'] = part_template % part_num
 
+    # NOTE(mjturek): With ppc64* nodes, partition images are expected to have
+    # a PrEP partition at the start of the disk. This is an 8 MiB partition
+    # with the boot and prep flags set. The bootloader should be installed
+    # here.
+    if (cpu_arch.startswith("ppc64") and boot_mode == "bios" and
+        boot_option == "local"):
+        LOG.debug("Add PReP boot partition (8 MB) to device: "
+                  "%(dev)s for node %(node)s",
+                  {'dev': dev, 'node': node_uuid})
+        boot_flag = 'boot' if disk_label == 'msdos' else None
+        part_num = dp.add_partition(8, part_type='primary',
+                                    boot_flag=boot_flag, extra_flags=['prep'])
+        part_dict['PReP Boot partition'] = part_template % part_num
     if ephemeral_mb:
         LOG.debug("Add ephemeral partition (%(size)d MB) to device: %(dev)s "
                   "for node %(node)s",
@@ -254,10 +278,10 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
               "for node %(node)s",
               {'dev': dev, 'size': root_mb, 'node': node_uuid})
 
-    boot_val = None
-    if (boot_mode == "bios" and boot_option == "local" and
-        disk_label == "msdos"):
-        boot_val = 'boot'
+    boot_val = 'boot' if (not cpu_arch.startswith("ppc64")
+                          and boot_mode == "bios"
+                          and boot_option == "local"
+                          and disk_label == "msdos") else None
 
     part_num = dp.add_partition(root_mb, boot_flag=boot_val)
 
@@ -321,10 +345,18 @@ def populate_image(src, dst):
 
 
 def block_uuid(dev):
-    """Get UUID of a block device."""
+    """Get UUID of a block device.
+
+    Try to fetch the UUID, if that fails, try to fetch the PARTUUID.
+    """
     out, _err = utils.execute('blkid', '-s', 'UUID', '-o', 'value', dev,
-                              run_as_root=True,
-                              check_exit_code=[0])
+                              run_as_root=True, check_exit_code=[0])
+    if not out:
+        LOG.debug('Falling back to partition UUID as the block device UUID '
+                  'was not found while examining %(device)s',
+                  {'device': dev})
+        out, _err = utils.execute('blkid', '-s', 'PARTUUID', '-o', 'value',
+                                  dev, run_as_root=True, check_exit_code=[0])
     return out.strip()
 
 
@@ -454,7 +486,7 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
 def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                  image_path, node_uuid, preserve_ephemeral=False,
                  configdrive=None, boot_option="netboot", boot_mode="bios",
-                 tempdir=None, disk_label=None):
+                 tempdir=None, disk_label=None, cpu_arch=""):
     """Create partitions and copy an image to the root partition.
 
     :param dev: Path for the device to work on.
@@ -477,6 +509,11 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
     :param disk_label: The disk label to be used when creating the
         partition table. Valid values are: "msdos", "gpt" or None; If None
         Ironic will figure it out according to the boot_mode parameter.
+    :param cpu_arch: Architecture of the node the disk device belongs to.
+        When using the default value of None, no architecture specific
+        steps will be taken. This default should be used for x86_64. When
+        set to ppc64*, architecture specific steps are taken for booting a
+        partition image locally.
     :returns: a dictionary containing the following keys:
         'root uuid': UUID of root partition
         'efi system partition uuid': UUID of the uefi system partition
@@ -505,7 +542,8 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                                     commit=commit,
                                     boot_option=boot_option,
                                     boot_mode=boot_mode,
-                                    disk_label=disk_label)
+                                    disk_label=disk_label,
+                                    cpu_arch=cpu_arch)
         LOG.info("Successfully completed the disk device"
                  " %(dev)s partitioning for node %(node)s",
                  {'dev': dev, "node": node_uuid})
@@ -520,7 +558,7 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                 _("Root device '%s' not found") % root_part)
 
         for part in ('swap', 'ephemeral', 'configdrive',
-                     'efi system partition'):
+                     'efi system partition', 'PReP Boot partition'):
             part_device = part_dict.get(part)
             LOG.debug("Checking for %(part)s device (%(dev)s) on node "
                       "%(node)s.", {'part': part, 'dev': part_device,
@@ -568,9 +606,14 @@ def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
 
     uuids_to_return = {
         'root uuid': root_part,
-        'efi system partition uuid': part_dict.get('efi system partition')
+        'efi system partition uuid': part_dict.get('efi system partition'),
     }
 
+    if cpu_arch.startswith('ppc'):
+        uuids_to_return[
+            'PReP Boot partition uuid'
+        ] = part_dict.get('PReP Boot partition')
+
     try:
         for part, part_dev in uuids_to_return.items():
             if part_dev:
@@ -627,7 +670,8 @@ def _get_labelled_partition(device_path, label, node_uuid):
               returns None.
     """
     try:
-        utils.execute('partprobe', device_path, run_as_root=True)
+        utils.execute('partprobe', device_path, run_as_root=True,
+                      attempts=CONF.disk_utils.partprobe_attempts)
 
         # lsblk command
         output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
@@ -714,6 +758,28 @@ def _fix_gpt_structs(device, node_uuid):
         raise exception.InstanceDeployFailure(msg)
 
 
+def fix_gpt_partition(device, node_uuid):
+    """Fix GPT partition
+
+    Fix GPT table information when image is written to a disk which
+    has a bigger extend (e.g. 30GB image written on a 60Gb physical disk).
+
+    :param device: The device path.
+    :param node_uuid: UUID of the Node.
+    :raises: InstanceDeployFailure if exception is caught.
+    """
+    try:
+        disk_is_gpt_partitioned = _is_disk_gpt_partitioned(device, node_uuid)
+        if disk_is_gpt_partitioned:
+            _fix_gpt_structs(device, node_uuid)
+    except Exception as e:
+        msg = (_('Failed to fix GPT partition on disk %(disk)s '
+                 'for node %(node)s. Error: %(error)s') %
+               {'disk': device, 'node': node_uuid, 'error': e})
+        LOG.error(msg)
+        raise exception.InstanceDeployFailure(msg)
+
+
 def create_config_drive_partition(node_uuid, device, configdrive):
     """Create a partition for config drive
 
@@ -746,6 +812,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                   "device: %(dev)s for node %(node)s",
                   {'dev': device, 'size': confdrive_mb, 'node': node_uuid})
 
+        fix_gpt_partition(device, node_uuid)
         if config_drive_part:
             LOG.debug("Configdrive for node %(node)s exists at "
                       "%(part)s",
@@ -754,7 +821,6 @@ def create_config_drive_partition(node_uuid, device, configdrive):
             cur_parts = set(part['number'] for part in list_partitions(device))
 
             if _is_disk_gpt_partitioned(device, node_uuid):
-                _fix_gpt_structs(device, node_uuid)
                 create_option = '0:-%dMB:0' % MAX_CONFIG_DRIVE_SIZE_MB
                 utils.execute('sgdisk', '-n', create_option, device,
                               run_as_root=True)
