@@ -26,6 +26,7 @@ import re
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
+from oslo_service import loopingcall
 from oslo_utils import excutils
 from oslo_utils import specs_matcher
 from oslo_utils import strutils
@@ -52,7 +53,7 @@ LOG = logging.getLogger(__name__)
 VALID_ROOT_DEVICE_HINTS = {
     'size': int, 'model': str, 'wwn': str, 'serial': str, 'vendor': str,
     'wwn_with_extension': str, 'wwn_vendor_extension': str, 'name': str,
-    'rotational': bool, 'hctl': str,
+    'rotational': bool, 'hctl': str, 'by_path': str,
 }
 
 
@@ -99,8 +100,8 @@ def execute(*cmd, **kwargs):
     LOG.debug('Execution completed, command line is "%s"',
               ' '.join(map(str, cmd)))
     if log_stdout:
-        LOG.debug('Command stdout is: "%s"' % result[0])
-    LOG.debug('Command stderr is: "%s"' % result[1])
+        LOG.debug('Command stdout is: "%s"', result[0])
+    LOG.debug('Command stderr is: "%s"', result[1])
     return result
 
 
@@ -348,6 +349,8 @@ def match_root_device_hints(devices, root_device_hints):
                      (not rotational).
         :hctl: (String): The SCSI address: Host, channel, target and lun.
                          For example: '1:0:0:0'.
+        :by_path: (String): The alternative device name,
+                  e.g. /dev/disk/by-path/pci-0000:00
 
     :param root_device_hints: A dictionary with the root device hints.
     :raises: ValueError, if some information is invalid.
@@ -415,3 +418,88 @@ def match_root_device_hints(devices, root_device_hints):
             return dev
 
     LOG.warning('No device found that matches the root device hints')
+
+
+def wait_for_disk_to_become_available(device):
+    """Wait for a disk device to become available.
+
+    Waits for a disk device to become available for use by
+    waiting until all process locks on the device have been
+    released.
+
+    Timeout and iteration settings come from the configuration
+    options used by the in-library disk_partitioner:
+    ``check_device_interval`` and ``check_device_max_retries``.
+
+    :params device: The path to the device.
+    :raises: IronicException If the disk fails to become
+        available.
+    """
+    retries = [0]
+    pids = ['']
+    stderr = ['']
+    interval = CONF.disk_partitioner.check_device_interval
+    max_retries = CONF.disk_partitioner.check_device_max_retries
+
+    def _wait_for_disk(device, retries, max_retries, pids, stderr):
+        # A regex is likely overkill here, but variations in fuser
+        # means we should likely use it.
+        fuser_pids_re = re.compile(r'\d+')
+
+        retries[0] += 1
+        if retries[0] > max_retries:
+            raise loopingcall.LoopingCallDone()
+
+        # There are 'psmisc' and 'busybox' versions of the 'fuser' program. The
+        # 'fuser' programs differ in how they output data to stderr.  The
+        # busybox version does not output the filename to stderr, while the
+        # standard 'psmisc' version does output the filename to stderr.  How
+        # they output to stdout is almost identical in that only the PIDs are
+        # output to stdout, with the 'psmisc' version adding a leading space
+        # character to the list of PIDs.
+        try:
+            # NOTE(ifarkas): fuser returns a non-zero return code if none of
+            #                the specified files is accessed.
+            # NOTE(TheJulia): fuser does not report LVM devices as in use
+            #                 unless the LVM device-mapper device is the
+            #                 device that is directly polled.
+            # NOTE(TheJulia): The -m flag allows fuser to reveal data about
+            #                 mounted filesystems, which should be considered
+            #                 busy/locked. That being said, it is not used
+            #                 because busybox fuser has a different behavior.
+            # NOTE(TheJuia): fuser outputs a list of found PIDs to stdout.
+            #                All other text is returned via stderr, and the
+            #                output to a terminal is merged as a result.
+            out, err = execute('fuser', device, check_exit_code=[0, 1],
+                               run_as_root=True)
+
+            if not out and not err:
+                raise loopingcall.LoopingCallDone()
+
+            stderr[0] = err
+            # NOTE: findall() returns a list of matches, or an empty list if no
+            # matches
+            pids[0] = fuser_pids_re.findall(out)
+
+        except processutils.ProcessExecutionError as exc:
+            LOG.warning('Failed to check the device %(device)s with fuser:'
+                        ' %(err)s', {'device': device, 'err': exc})
+
+    timer = loopingcall.FixedIntervalLoopingCall(
+        _wait_for_disk,
+        device, retries, max_retries, pids, stderr)
+    timer.start(interval=interval).wait()
+
+    if retries[0] > max_retries:
+        if pids[0]:
+            raise exception.IronicException(
+                _('Processes with the following PIDs are holding '
+                  'device %(device)s: %(pids)s. '
+                  'Timed out waiting for completion.')
+                % {'device': device, 'pids': ', '.join(pids[0])})
+        else:
+            raise exception.IronicException(
+                _('Fuser exited with "%(fuser_err)s" while checking '
+                  'locks for device %(device)s. Timed out waiting for '
+                  'completion.')
+                % {'device': device, 'fuser_err': stderr[0]})
